package es.ucm.fdi.grasia.faerie.container.internal.local;

import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.felix.scr.annotations.Activate;
import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Deactivate;
import org.apache.felix.scr.annotations.Reference;
import org.apache.felix.scr.annotations.ReferenceCardinality;
import org.apache.felix.scr.annotations.ReferencePolicy;
import org.apache.felix.scr.annotations.Service;
import org.apache.log4j.Logger;

import es.ucm.fdi.grasia.faerie.container.ContextContainer;
import es.ucm.fdi.grasia.faerie.container.ContextContainerService;
import es.ucm.fdi.grasia.faerie.core.*;
import es.ucm.fdi.grasia.faerie.core.model.BundleEntity;
import es.ucm.fdi.grasia.faerie.core.util.ContextURIs;
import java.util.Map.Entry;
import java.util.concurrent.*;

/**
 *
 * @author Chema
 */
@Service
@Component(immediate = true)
@Reference(name = "contextObserver", referenceInterface = ContextObserver.class, policy = ReferencePolicy.DYNAMIC,
cardinality = ReferenceCardinality.OPTIONAL_MULTIPLE, target = "(&(!(service.imported=*))(!(service.uri=*)))")
@SuppressWarnings({ "unchecked", "rawtypes" })
public class LocalContextContainer implements ContextContainer, ContextContainerService {

	public static final long THREAD_WAIT = 10000;
	private final Map<URI, Entity> entities = setUpEntities();
	private final Map<URI, Attribute> attributeInstances = setUpAttributeInstances();
	private final Map<URI, Relationship> relationshipInstances = setUpRelationshipInstances();
	private final Set<ContextObserver> contextObservers = setUpContextObservers();
	private final Map<URI, AlternativeObserver> alternativeObservers = setUpAlternativeObservers();
	private final Map<UUID, Set<URI>> requestersElements = setUpRequestersElements();
	private final Map<URI, Set<UUID>> elementsRequesters = setUpElementsRequesters();
	private final ScheduledExecutorService executorService = setUpExecutorService();

	protected Map<URI, Entity> setUpEntities() {
		return new HashMap<URI, Entity>();
	}

	protected Map<URI, Attribute> setUpAttributeInstances() {
		return new HashMap<URI, Attribute>();
	}

	protected Map<URI, Relationship> setUpRelationshipInstances() {
		return new HashMap<URI, Relationship>();
	}

	protected Set<ContextObserver> setUpContextObservers() {
		return new LinkedHashSet<ContextObserver>();
	}

	protected Map<URI, AlternativeObserver> setUpAlternativeObservers() {
		return new ConcurrentHashMap<URI, AlternativeObserver>();
	}

	protected Map<UUID, Set<URI>> setUpRequestersElements() {
		return new HashMap<UUID, Set<URI>>();
	}

	protected Map<URI, Set<UUID>> setUpElementsRequesters() {
		return new HashMap<URI, Set<UUID>>();
	}

	protected Set<URI> setUpIdSet() {
		return new HashSet<URI>();
	}

	protected Set<UUID> setUpUUIDSet() {
		return new HashSet<UUID>();
	}

	protected ScheduledExecutorService setUpExecutorService() {
		return FaerieExecutorService.getExecutorService();
	}

	protected class AlternativeObserver extends ContextEventAdapter {

		public static final long DELAY = 1000;
		private ContextElement element;
		private Map<URI, ContextObserver> alternatives = setUpAlternatives();
		private URI current = null;
		private ScheduledFuture future = null;

		protected Map<URI, ContextObserver> setUpAlternatives() {
			return new HashMap<URI, ContextObserver>();
		}

		public AlternativeObserver(ContextElement element) {
			this.element = element;
		}

		public synchronized void addAlternative(ContextObserver observer) {
			if (observer instanceof AbstractContextComponent) {
				BundleEntity obsEnt = ((AbstractContextComponent) observer).getBundleEntity();
				URI obsId = obsEnt.getId();
				if (alternatives.put(obsId, observer) == null && current == null) {
					Logger.getLogger(this.getClass()).trace("Add " + obsId + " as alt to " + element);
					observer.elementAdded(createContextEvent(ContextEvent.ELEMENT_ADDED, element));
				}
			}
		}

		public synchronized void removeAlternative(ContextObserver observer) {
			if (observer instanceof AbstractContextComponent) {
				BundleEntity obsEnt = ((AbstractContextComponent) observer).getBundleEntity();
				URI obsId = obsEnt.getId();
				if (alternatives.remove(obsId) != null)
					if (current == null) {
						Logger.getLogger(this.getClass()).trace("Rem " + obsId + " as alt to " + element);
						observer.elementRemoved(createContextEvent(ContextEvent.ELEMENT_REMOVED, element));
					} else if (obsId.equals(current)) {
						Logger.getLogger(this.getClass()).trace("Rem " + obsId + " as curr to " + element);
						observer.elementRemoved(createContextEvent(ContextEvent.ELEMENT_REMOVED, element));
						ContextEvent event = createContextEvent(ContextEvent.ELEMENT_ADDED, element);
						for (Entry<URI, ContextObserver> entry : alternatives.entrySet()) {
							Logger.getLogger(this.getClass()).trace("Put " + entry.getKey() + " as alt to "
									+ element);
							entry.getValue().elementAdded(event);
						}
						current = null;
					}
			}
		}

		public synchronized int countAlternatives() {
			return alternatives.size();
		}

		@Override
		public synchronized void elementUpdated(ContextEvent event) {
			Attribute source = event.getSourceAttributeInstance();
			if (source.isRemotelyUpdateable())
				return;
			if (element.equals(source)) {
				ContextValue value = source.getValue();
				if (value != null) {
					Entity obsEnt = value.getSource();
					final URI obsId = obsEnt.getId();
					Object data = value.getData();
					if (future != null && obsId.equals(current) && data != null) {
						future.cancel(false);
						Logger.getLogger(this.getClass()).trace("Canceled " + future);
						future = null;
					} else if (obsId.equals(current) && data == null) {
						future = executorService.schedule(new Runnable() {

							@Override
							public void run() {
								synchronized (AlternativeObserver.this) {
									Logger.getLogger(this.getClass()).trace("Executing future\tcurrent = " + current);
									ContextEvent newEvent = createContextEvent(ContextEvent.ELEMENT_ADDED, element);
									for (Entry<URI, ContextObserver> entry : alternatives.entrySet())
										if (!entry.getKey().equals(current)) {
											Logger.getLogger(this.getClass()).trace("Put " + entry.getKey()
													+ " as alt to " + element);
											entry.getValue().elementAdded(newEvent);
										} else
											Logger.getLogger(this.getClass()).trace("Quit " + entry.getKey()
													+ " as curr to " + element);
									current = null;
									future = null;
								}
							}
						}, DELAY, TimeUnit.MILLISECONDS);
						Logger.getLogger(this.getClass()).trace("Scheduled " + future);
					} else if (current == null && data != null) {
						current = obsId;
						ContextEvent newEvent = createContextEvent(ContextEvent.ELEMENT_REMOVED, element);
						Logger.getLogger(this.getClass()).trace("current = " + current);
						for (Entry<URI, ContextObserver> entry : alternatives.entrySet())
							if (!entry.getKey().equals(current)) {
								Logger.getLogger(this.getClass()).trace("Quit " + entry.getKey() + " as alt to "
										+ element);
								entry.getValue().elementRemoved(newEvent);
							} else
								Logger.getLogger(this.getClass()).trace("Put " + entry.getKey() + " as curr to "
										+ element);
					}
				}
			}
		}
	}

	@Activate
	@Override
	public void activate() {
		Logger.getLogger(this.getClass()).debug("(" + ContextContainer.class.getSimpleName() + ") Activating * " + this);
	}

	@Deactivate
	@Override
	public void deactivate() {
		Logger.getLogger(this.getClass()).debug(
				"(" + ContextContainer.class.getSimpleName() + ") Deactivating * " + this);
	}

	protected synchronized Map<URI, Entity> getAllEntities() {
		Map<URI, Entity> clone = setUpEntities();
		clone.putAll(entities);
		return clone;
	}

	protected synchronized Map<URI, Attribute> getAllAttributeInstances() {
		Map<URI, Attribute> clone = setUpAttributeInstances();
		clone.putAll(attributeInstances);
		return clone;
	}

	protected synchronized Map<URI, Relationship> getAllRelationshipInstances() {
		Map<URI, Relationship> clone = setUpRelationshipInstances();
		clone.putAll(relationshipInstances);
		return clone;
	}

	protected static ContextEvent createContextEvent(int eventType, Object arg) {
		return new ContextEvent(null, eventType, arg);
	}

	protected class AddContextObserverTask implements Runnable {

		private final ContextObserver observer;

		protected AddContextObserverTask(ContextObserver contextObserver) {
			this.observer = contextObserver;
		}

		@Override
		public void run() {
			synchronized (observer) {
				Logger.getLogger(LocalContextContainer.this.getClass()).debug(
						"Adding " + ContextObserver.class.getSimpleName() + ": " + observer + " * "
						+ LocalContextContainer.this);

				Map<URI, Entity> entitiesCopy;
				Map<URI, Attribute> attributeInstancesCopy;
				Map<URI, Relationship> relationshipInstancesCopy;
				synchronized (LocalContextContainer.this) {
					contextObservers.add(observer);

					entitiesCopy = getAllEntities();
					attributeInstancesCopy = getAllAttributeInstances();
					relationshipInstancesCopy = getAllRelationshipInstances();
				}

				for (Entity entity : entitiesCopy.values()) {
					ContextEvent event = createContextEvent(ContextEvent.ELEMENT_ADDED, entity);
					if (observer.handle(event) != 0)
						observer.elementAdded(event);
				}
				for (final Attribute attributeInstance : attributeInstancesCopy.values()) {
					ContextEvent event = createContextEvent(ContextEvent.ELEMENT_ADDED, attributeInstance);
					if ((observer.handle(event) & ContextObserver.HANDLE_WRITE) != 0) {
						AlternativeObserver alter = alternativeObservers.get(attributeInstance.getId());
						if (alter == null) {
							alternativeObservers.put(attributeInstance.getId(), alter = new AlternativeObserver(
									attributeInstance));
							attributeInstance.addObserver(alter);
						}
						alter.addAlternative(observer);
					} else if (observer.handle(event) == ContextObserver.HANDLE_READ)
						observer.elementAdded(event);
				}
				for (Relationship relationshipInstance : relationshipInstancesCopy.values()) {
					ContextEvent event = createContextEvent(ContextEvent.ELEMENT_ADDED, relationshipInstance);
					if (observer.handle(event) != 0)
						observer.elementAdded(event);
				}
			}
		}
	}

	protected Runnable createAddContextObserverTask(ContextObserver contextObserver) {
		return new AddContextObserverTask(contextObserver);
	}

	@Override
	public void bindContextObserver(ContextObserver contextObserver) {
		executorService.submit(createAddContextObserverTask(contextObserver));
	}

	protected class RemoveContextObserverTask implements Runnable {

		private final ContextObserver observer;

		protected RemoveContextObserverTask(ContextObserver contextObserver) {
			this.observer = contextObserver;
		}

		@Override
		public void run() {
			synchronized (observer) {
				Logger.getLogger(LocalContextContainer.this.getClass()).debug(
						"Removing " + ContextObserver.class.getSimpleName() + ": " + observer + " * "
						+ LocalContextContainer.this);

				Map<URI, Relationship> relationshipInstancesCopy;
				Map<URI, Attribute> attributeInstancesCopy;
				Map<URI, Entity> entitiesCopy;
				synchronized (LocalContextContainer.this) {
					contextObservers.remove(observer);

					entitiesCopy = getAllEntities();
					relationshipInstancesCopy = getAllRelationshipInstances();
					attributeInstancesCopy = getAllAttributeInstances();
				}

				for (Relationship relationshipInstance : relationshipInstancesCopy.values()) {
					ContextEvent event = createContextEvent(ContextEvent.ELEMENT_ADDED, relationshipInstance);
					if (observer.handle(event) != 0)
						observer.elementRemoved(event);
				}
				for (final Attribute attributeInstance : attributeInstancesCopy.values()) {
					ContextEvent event = createContextEvent(ContextEvent.ELEMENT_REMOVED, attributeInstance);
					if ((observer.handle(event) & ContextObserver.HANDLE_WRITE) != 0) {
						AlternativeObserver alter = alternativeObservers.get(attributeInstance.getId());
						alter.removeAlternative(observer);
						if (alter.countAlternatives() == 0) {
							attributeInstance.removeObserver(alter);
							alternativeObservers.remove(attributeInstance.getId());
						}
					} else if (observer.handle(event) == ContextObserver.HANDLE_READ)
						observer.elementRemoved(event);
				}
				for (Entity entity : entitiesCopy.values()) {
					ContextEvent event = createContextEvent(ContextEvent.ELEMENT_REMOVED, entity);
					if (observer.handle(event) != 0)
						observer.elementRemoved(event);
				}
			}
		}
	}

	protected Runnable createRemoveContextObserverTask(ContextObserver contextObserver) {
		return new RemoveContextObserverTask(contextObserver);
	}

	@Override
	public void unbindContextObserver(ContextObserver contextObserver) {
		Future future = executorService.submit(createRemoveContextObserverTask(contextObserver));
		try {
			future.get(THREAD_WAIT, TimeUnit.MILLISECONDS);
		} catch (ExecutionException ex) {
		} catch (InterruptedException ex) {
			future.cancel(true);
		} catch (TimeoutException ex) {
			future.cancel(true);
		}
	}

	protected class FireElementAddedTask implements Runnable {

		private final ContextObserver observer;
		private final ContextElement element;

		protected FireElementAddedTask(ContextObserver contextObserver, ContextElement contextElement) {
			this.observer = contextObserver;
			this.element = contextElement;
		}

		@Override
		public void run() {
			synchronized (observer) {
				ContextEvent event = createContextEvent(ContextEvent.ELEMENT_ADDED, element);
				if ((observer.handle(event) & ContextObserver.HANDLE_WRITE) != 0) {
					AlternativeObserver alter = alternativeObservers.get(element.getId());
					if (alter == null) {
						alternativeObservers.put(element.getId(), alter = new AlternativeObserver(element));
						element.addObserver(alter);
					}
					alter.addAlternative(observer);
				} else if (observer.handle(event) == ContextObserver.HANDLE_READ)
					observer.elementAdded(event);
			}
		}
	}

	protected Runnable createFireElementAddedTask(ContextObserver contextObserver, ContextElement contextElement) {
		return new FireElementAddedTask(contextObserver, contextElement);
	}

	protected synchronized void includeRelationshipInstance(Relationship relationshipInstance) {
		Logger.getLogger(this.getClass()).debug(
				"Adding " + Relationship.class.getSimpleName() + ": " + relationshipInstance + " * " + this);

		relationshipInstances.put(relationshipInstance.getId(), relationshipInstance);

		for (ContextObserver observer : contextObservers)
			executorService.submit(createFireElementAddedTask(observer, relationshipInstance));
	}

	protected synchronized void includeAttributeInstance(Attribute attributeInstance) {
		Logger.getLogger(this.getClass()).debug(
				"Adding " + Attribute.class.getSimpleName() + ": " + attributeInstance + " * " + this);

		attributeInstances.put(attributeInstance.getId(), attributeInstance);

		for (ContextObserver observer : contextObservers)
			executorService.submit(createFireElementAddedTask(observer, attributeInstance));
	}

	protected synchronized void includeEntity(Entity entity) {
		Logger.getLogger(this.getClass()).debug("Adding " + Entity.class.getSimpleName() + ": " + entity + " * " + this);

		entities.put(entity.getId(), entity);

		for (ContextObserver observer : contextObservers)
			executorService.submit(createFireElementAddedTask(observer, entity));
	}

	protected class FireElementRemovedTask implements Runnable {

		private final ContextObserver observer;
		private final ContextElement element;

		protected FireElementRemovedTask(ContextObserver contextObserver, ContextElement contextElement) {
			this.observer = contextObserver;
			this.element = contextElement;
		}

		@Override
		public void run() {
			synchronized (observer) {
				ContextEvent event = createContextEvent(ContextEvent.ELEMENT_REMOVED, element);
				if ((observer.handle(event) & ContextObserver.HANDLE_WRITE) != 0) {
					AlternativeObserver alter = alternativeObservers.get(element.getId());
					alter.removeAlternative(observer);
					if (alter.countAlternatives() == 0) {
						element.removeObserver(alter);
						alternativeObservers.remove(element.getId());
					}
				} else if (observer.handle(event) == ContextObserver.HANDLE_READ)
					observer.elementRemoved(event);
			}
		}
	}

	protected Runnable createFireElementRemovedTask(ContextObserver contextObserver, ContextElement contextElement) {
		return new FireElementRemovedTask(contextObserver, contextElement);
	}

	protected synchronized void excludeRelationshipInstance(Relationship relationshipInstance) {
		Logger.getLogger(this.getClass()).debug(
				"Removing " + Relationship.class.getSimpleName() + ": " + relationshipInstance + " * " + this);

		for (ContextObserver observer : contextObservers)
			executorService.submit(createFireElementRemovedTask(observer, relationshipInstance));

		relationshipInstances.remove(relationshipInstance.getId());
	}

	protected synchronized void excludeAttributeInstance(Attribute attributeInstance) {
		Logger.getLogger(this.getClass()).debug(
				"Removing " + Attribute.class.getSimpleName() + ": " + attributeInstance + " * " + this);

		for (ContextObserver observer : contextObservers)
			executorService.submit(createFireElementRemovedTask(observer, attributeInstance));

		attributeInstances.remove(attributeInstance.getId());
	}

	protected synchronized void excludeEntity(Entity entity) {
		Logger.getLogger(this.getClass()).debug(
				"Removing " + Entity.class.getSimpleName() + ": " + entity + " * " + this);

		for (ContextObserver observer : contextObservers)
			executorService.submit(createFireElementRemovedTask(observer, entity));

		entities.remove(entity.getId());
	}

	protected synchronized void bindRequester(UUID requesterId, URI elementId) {
		Set<URI> elements = requestersElements.get(requesterId);
		if (elements == null)
			requestersElements.put(requesterId, elements = setUpIdSet());
		elements.add(elementId);
		Set<UUID> requesters = elementsRequesters.get(elementId);
		if (requesters == null)
			elementsRequesters.put(elementId, requesters = setUpUUIDSet());
		requesters.add(requesterId);
	}

	protected synchronized void unbindRequester(UUID requesterId, URI elementId) {
		Set<URI> elements = requestersElements.get(requesterId);
		if (elements != null)
			elements.remove(elementId);
		Set<UUID> requesters = elementsRequesters.get(elementId);
		if (requesters != null)
			requesters.remove(requesterId);
	}

	protected synchronized int countRequesters(URI elementId) {
		Set<UUID> requesters = elementsRequesters.get(elementId);
		if (requesters == null)
			return 0;
		return requesters.size();
	}

	@Override
	public synchronized boolean contains(URI elementId, UUID requesterId) {
		return peek(elementId, requesterId) != null;
	}

	@Override
	public synchronized <T extends ContextElement> T peek(URI elementId, UUID requesterId) {
		if (elementId.getQuery() != null)
			return (T) relationshipInstances.get(elementId);
		else if (elementId.getFragment() != null)
			return (T) attributeInstances.get(elementId);
		else
			return (T) entities.get(elementId);
	}

	@Override
	public synchronized <T extends ContextElement> T request(T element, UUID requesterId) {
		Logger.getLogger(this.getClass()).trace("(" + element + ", " + requesterId + ") * " + this);

		URI elementId = element.getId();
		if (element instanceof Entity) {
			Entity entity = entities.get(elementId);
			if (entity == null)
				includeEntity(entity = (Entity) element);
			bindRequester(requesterId, elementId);

			return (T) entity;
		} else if (element instanceof Attribute) {
			Attribute ai = attributeInstances.get(elementId);
			if (ai == null) {
				ai = (Attribute) element;
				Entity ent = ai.getEntity();
				Entity existing = entities.get(ent.getId());
				if (existing == null)
					includeEntity(existing = ent);
				else
					ai.setEntity(existing);
				includeAttributeInstance(ai);
				existing.putAttributeInstance(ai);
			}
			bindRequester(requesterId, elementId);

			return (T) ai;
		} else if (element instanceof Relationship) {
			Relationship ri = relationshipInstances.get(elementId);
			if (ri == null) {
				ri = (Relationship) element;
				Entity sourceEnt = ri.getSourceEntity();
				Entity existingSource = entities.get(sourceEnt.getId());
				if (existingSource == null)
					includeEntity(existingSource = sourceEnt);
				else
					ri.setSourceEntity(existingSource);
				List<Entity> targetEnts = ri.getAllTargetEntities();
				for (int i = 0; i < targetEnts.size(); ++i) {
					Entity targetEnt = targetEnts.get(i);
					if (targetEnt != null) {
						Entity existingTarget = entities.get(targetEnt.getId());
						if (existingTarget == null)
							includeEntity(targetEnt);
						else
							ri.setTargetEntity(i, existingTarget);
					}
				}
				includeRelationshipInstance(ri);
				existingSource.putRelationshipInstance(ri);
			}
			bindRequester(requesterId, elementId);

			return (T) ri;
		} else
			return null;
	}

	@Override
	public synchronized void release(URI elementId, UUID requesterId) {
		Logger.getLogger(this.getClass()).trace("(" + elementId + ", " + requesterId + ") * " + this);

		if (elementId.getQuery() != null) {
			Relationship ri = relationshipInstances.get(elementId);
			if (ri != null) {
				unbindRequester(requesterId, elementId);
				if (countRequesters(elementId) == 0) {
					URI sourceEntityId = ri.getSourceEntity().getId();
					Entity sourceEntity = entities.get(sourceEntityId);
					sourceEntity.removeRelationshipInstance(ri.getClass());
					excludeRelationshipInstance(ri);
					List<Entity> targetEnts = ri.getAllTargetEntities();
					for (int i = 0; i < targetEnts.size(); ++i) {
						Entity targetEnt = targetEnts.get(i);
						if (targetEnt != null) {
							URI targetEntId = targetEnt.getId();
							targetEnt = entities.get(targetEntId);
							if (targetEnt != null && targetEnt.getAllRelationshipsInstances().isEmpty()
									&& targetEnt.getAllAttributesInstances().isEmpty()
									&& countRequesters(targetEntId) == 0)
								excludeEntity(targetEnt);
						}
					}
					if (sourceEntity.getAllRelationshipsInstances().isEmpty()
							&& sourceEntity.getAllAttributesInstances().isEmpty()
							&& countRequesters(sourceEntityId) == 0)
						excludeEntity(sourceEntity);
				}
			}
		} else if (elementId.getFragment() != null) {
			Attribute ai = attributeInstances.get(elementId);
			if (ai != null) {
				unbindRequester(requesterId, elementId);
				if (countRequesters(elementId) == 0) {
					URI entityId = ai.getEntity().getId();
					Entity entity = entities.get(entityId);
					entity.removeAttributeInstance(ai.getClass());
					excludeAttributeInstance(ai);
					if (entity.getAllRelationshipsInstances().isEmpty() && entity.getAllAttributesInstances().isEmpty()
							&& countRequesters(entityId) == 0)
						excludeEntity(entity);
				}
			}
		} else {
			Entity entity = entities.get(elementId);
			if (entity != null) {
				unbindRequester(requesterId, elementId);
				if (entity.getAllRelationshipsInstances().isEmpty() && entity.getAllAttributesInstances().isEmpty()
						&& countRequesters(elementId) == 0)
					excludeEntity(entity);
			}
		}
	}

	@Override
	public synchronized void releaseAll(UUID requesterId) {
		Logger.getLogger(this.getClass()).trace("(" + requesterId + ") * " + this);

		Set<URI> elements = requestersElements.get(requesterId);
		if (elements != null) {
			Set<URI> clone = setUpIdSet();
			clone.addAll(elements);
			for (URI elementId : clone)
				release(elementId, requesterId);
		}
	}

	@Override
	public String toString() {
		return ContextURIs.localAuthority().toString();
	}
}
